@@ -6,7 +6,7 @@ packeting module provides classes for Raet packets
 '''
 
 # Import python libs
-from collections import Mapping
+from collections import Mapping, deque
 try:
     import simplejson as json
 except ImportError:
@@ -24,7 +24,10 @@ from ioflo.base.aiding import packByte, unpackByte
 from ioflo.base.consoling import getConsole
 console = getConsole()
 
+# Import raet libs
+from ..abiding import *  # import globals
 from .. import raeting
+from ..raeting import PcktKind, TailSize, CoatKind, FootSize, FootKind, BodyKind, HeadKind
 
 class Part(object):
     '''
@@ -37,7 +40,7 @@ class Part(object):
         Setup Part instance
         '''
         self.packet = packet  # Packet this Part belongs too
-        self.packed = ''
+        self.packed = b''
 
     def __len__(self):
         '''
@@ -71,28 +74,26 @@ class TxHead(Head):
         '''
         Composes .packed, which is the packed form of this part
         '''
-        self.packed = ''
+        self.packed = b''
         data = self.packet.data  # for speed
         data['fl'] = self.packet.foot.size
         data['fg'] = "{0:02x}".format(self.packFlags())
 
-        # kit always includes raet id, packet length, and header kind fields
-        kit = odict([('ri', 'RAET'), ('pl', 0), ('hl', 0)])
+        # kit always includes raet id, packet length, header kind and flag fields
+        kit = odict([('ri', 'RAET'), ('pl', 0), ('hl', 0), ('fg', '00')])
         for k, v in raeting.PACKET_DEFAULTS.items():  # include if not equal to default
             if ((k in raeting.PACKET_HEAD_FIELDS) and
                 (k not in raeting.PACKET_FLAGS) and
                 (data[k] != v)):
                 kit[k] = data[k]
 
-        if data['hk'] == raeting.headKinds.raet:
-            packed = ''
+        if data['hk'] == HeadKind.raet:
+            packed = b''
             lines = []
             for k, v in kit.items():
                 lines.append("{key} {val:{fmt}}".format(
                         key=k, val=v, fmt=raeting.PACKET_FIELD_FORMATS[k]))
-
-            packed = "\n".join(lines)
-            packed = '{0}{1}'.format(packed, raeting.HEAD_END)
+            packed = ns2b('\n'.join(lines)) + raeting.HEAD_END
             hl = len(packed)
             if hl > raeting.MAX_HEAD_SIZE:
                 emsg = "Head length of {0}, exceeds max of {1}".format(hl, MAX_HEAD_SIZE)
@@ -108,27 +109,27 @@ class TxHead(Head):
             # Tray checks for packet length greater than UDP_MAX_PACKET_SIZE
             # and segments appropriately so pl may be truncated below in this case
             # substitute true lengths
-            packed = packed.replace('\npl {val:{fmt}}\n'.format(
+            packed = packed.replace(ns2b('\npl {val:{fmt}}\n'.format(
                                         val=kit['pl'],
-                                        fmt=raeting.PACKET_FIELD_FORMATS['pl']),
-                                    '\npl {0}\n'.format("{val:{fmt}}".format(
+                                        fmt=raeting.PACKET_FIELD_FORMATS['pl'])),
+                                    ns2b('\npl {0}\n'.format("{val:{fmt}}".format(
                                         val=pl,
-                                        fmt=raeting.PACKET_FIELD_FORMATS['pl'])[-4:]),
+                                        fmt=raeting.PACKET_FIELD_FORMATS['pl'])[-4:])),
                                     1)
-            packed = packed.replace('\nhl {val:{fmt}}\n'.format(
+            packed = packed.replace(ns2b('\nhl {val:{fmt}}\n'.format(
                                         val=kit['hl'],
-                                        fmt=raeting.PACKET_FIELD_FORMATS['hl']),
-                                    '\nhl {0}\n'.format("{val:{fmt}}".format(
+                                        fmt=raeting.PACKET_FIELD_FORMATS['hl'])),
+                                    ns2b('\nhl {0}\n'.format("{val:{fmt}}".format(
                                             val=hl,
-                                            fmt=raeting.PACKET_FIELD_FORMATS['hl'])[-2:]),
+                                            fmt=raeting.PACKET_FIELD_FORMATS['hl'])[-2:])),
                                     1)
             self.packed = packed
 
-        elif data['hk'] == raeting.headKinds.json:
+        elif data['hk'] == HeadKind.json:
             kit['pl'] = '0000000'  # need hex string so fixed length and jsonable
             kit['hl'] = '00'  # need hex string so fixed length and jsonable
-            packed = json.dumps(kit, separators=(',', ':'), encoding='ascii',)
-            packed = '{0}{1}'.format(packed, raeting.JSON_END)
+            packed = (ns2b(json.dumps(kit, separators=(',', ':'),
+                                     encoding='ascii',)) + raeting.JSON_END)
             hl = len(packed)
             if hl > raeting.MAX_HEAD_SIZE:
                 emsg = "Head length of {0}, exceeds max of {1}".format(hl, MAX_HEAD_SIZE)
@@ -142,8 +143,12 @@ class TxHead(Head):
             pl = hl + self.packet.coat.size + data['fl']
             data['pl'] = pl
             #substitute true length converted to 2 byte hex string
-            packed = packed.replace('"pl":"0000000"', '"pl":"{0}"'.format("{0:07x}".format(pl)[-7:]), 1)
-            self.packed = packed.replace('"hl":"00"', '"hl":"{0}"'.format("{0:02x}".format(hl)[-2:]), 1)
+            packed = packed.replace(ns2b('"pl":"0000000"'),
+                                    ns2b('"pl":"{0}"'.format("{0:07x}".format(pl)[-7:])),
+                                    1)  # JSON needs double quotes on strings
+            self.packed = packed.replace(ns2b('"hl":"00"'),
+                                         ns2b('"hl":"{0}"'.format("{0:02x}".format(hl)[-2:])),
+                                         1)  # JSON needs double quotes on strings
 
     def packFlags(self):
         '''
@@ -152,7 +157,7 @@ class TxHead(Head):
         values = []
         for field in raeting.PACKET_FLAG_FIELDS:
             values.append(1 if self.packet.data.get(field, 0) else 0)
-        return packByte(fmt='11111111', fields=values)
+        return packByte(fmt=b'11111111', fields=values)
 
 class RxHead(Head):
     '''
@@ -164,16 +169,16 @@ class RxHead(Head):
         .packet.data
         Raises PacketError if failure occurs
         '''
-        self.packed = ''
+        self.packed = b''
         data = self.packet.data  # for speed
         packed = self.packet.packed  # for speed
 
-        if packed.startswith('ri RAET\n') and raeting.HEAD_END in packed: # raet head
-            hk = raeting.headKinds.raet
+        if packed.startswith(b'ri RAET\n') and raeting.HEAD_END in packed: # raet head
+            hk = HeadKind.raet.value
             front, sep, back = packed.partition(raeting.HEAD_END)
-            self.packed = "{0}{1}".format(front, sep)
+            self.packed = front + sep
             kit = odict()
-            lines = front.split('\n')
+            lines = str(front.decode(encoding='ISO-8859-1')).split('\n')
             for line in lines:
                 key, val = line.split(' ')
                 if key not in raeting.PACKET_HEAD_FIELDS:
@@ -206,11 +211,11 @@ class RxHead(Head):
                 raise raeting.PacketError(emsg)
 
 
-        elif packed.startswith('{"ri":"RAET",') and raeting.JSON_END in packed: # json head
-            hk = raeting.headKinds.json
+        elif packed.startswith(ns2b('{"ri":"RAET",')) and raeting.JSON_END in packed: # json head
+            hk = HeadKind.json.value
             front, sep, back = packed.partition(raeting.JSON_END)
-            self.packed = "{0}{1}".format(front, sep)
-            kit = json.loads(front,
+            self.packed = front + sep
+            kit = json.loads(front.decode(encoding='ascii'),
                              encoding='ascii',
                              object_pairs_hook=odict)
             data.update(kit)
@@ -233,7 +238,7 @@ class RxHead(Head):
             data['pl'] = pl
 
         else:  # notify unrecognizable packet head
-            data['hk'] = raeting.headKinds.unknown
+            data['hk'] = HeadKind.unknown.value
             emsg = "Unrecognizable packet head."
             raise raeting.PacketError(emsg)
 
@@ -241,7 +246,7 @@ class RxHead(Head):
         '''
         Unpacks all the flag fields from a single two char hex string
         '''
-        values = unpackByte(fmt='11111111', byte=int(flags, 16), boolean=True)
+        values = unpackByte(fmt=b'11111111', byte=int(flags, 16), boolean=True)
         for i, field in enumerate(raeting.PACKET_FLAG_FIELDS):
             if field in self.packet.data:
                 self.packet.data[field] = values[i]
@@ -268,18 +273,21 @@ class TxBody(Body):
         '''
         Composes .packed, which is the packed form of this part
         '''
-        self.packed = ''
+        self.packed = b''
         bk = self.packet.data['bk']
-        if bk == raeting.bodyKinds.json:
+        if bk == BodyKind.json:
             if self.data:
-                self.packed = json.dumps(self.data, separators=(',', ':'))
-        elif bk == raeting.bodyKinds.msgpack:
+                self.packed = ns2b(json.dumps(self.data,
+                                              separators=(',', ':'),
+                                              encoding='utf-8'))
+        elif bk == BodyKind.msgpack:
             if self.data:
                 if not msgpack:
                     emsg = "Msgpack not installed."
                     raise raeting.PacketError(emsg)
-                self.packed = msgpack.dumps(self.data)
-        elif bk == raeting.bodyKinds.raw:
+                self.packed = msgpack.dumps(self.data,
+                                            encoding='utf-8')
+        elif bk == BodyKind.raw:
             self.packed = self.data # data is already formatted string
 
 class RxBody(Body):
@@ -293,33 +301,37 @@ class RxBody(Body):
         '''
         bk = self.packet.data['bk']
 
-        if bk not in raeting.BODY_KIND_NAMES:
-            self.packet.data['bk']= raeting.bodyKinds.unknown
+        if bk not in list(BodyKind):
+            self.packet.data['bk']= BodyKind.unknown.value
             emsg = "Unrecognizable packet body."
             raise raeting.PacketError(emsg)
 
         self.data = odict()
 
-        if bk == raeting.bodyKinds.json:
+        if bk == BodyKind.json:
             if self.packed:
-                kit = json.loads(self.packed, object_pairs_hook=odict)
+                kit = json.loads(self.packed.decode(encoding='utf-8'),
+                                 object_pairs_hook=odict,
+                                 encoding='utf-8')
                 if not isinstance(kit, Mapping):
                     emsg = "Packet body not a mapping."
                     raise raeting.PacketError(emsg)
                 self.data = kit
-        elif bk == raeting.bodyKinds.msgpack:
+        elif bk == BodyKind.msgpack:
             if self.packed:
                 if not msgpack:
                     emsg = "Msgpack not installed."
                     raise raeting.PacketError(emsg)
-                kit = msgpack.loads(self.packed, object_pairs_hook=odict)
+                kit = msgpack.loads(self.packed,
+                                    object_pairs_hook=odict,
+                                    encoding='utf-8')
                 if not isinstance(kit, Mapping):
                     emsg = "Packet body not a mapping."
                     raise raeting.PacketError(emsg)
                 self.data = kit
-        elif bk == raeting.bodyKinds.raw:
-            self.data = self.packed # return as string
-        elif bk == raeting.bodyKinds.nada:
+        elif bk == BodyKind.raw:
+            self.data = self.packed # return as bytes
+        elif bk == BodyKind.nada:
             pass
 
 class Coat(Part):
@@ -339,16 +351,16 @@ class TxCoat(Coat):
         '''
         Composes .packed, which is the packed form of this part
         '''
-        self.packed = ''
+        self.packed = b''
         ck = self.packet.data['ck']
 
-        if ck == raeting.coatKinds.nacl:
+        if ck == CoatKind.nacl:
             msg = self.packet.body.packed
             if msg:
                 cipher, nonce = self.packet.encrypt(msg)
-                self.packed = "".join([cipher, nonce])
+                self.packed = b''.join([cipher, nonce])
 
-        if ck == raeting.coatKinds.nada:
+        if ck == CoatKind.nada:
             self.packed = self.packet.body.packed
 
 class RxCoat(Coat):
@@ -361,14 +373,14 @@ class RxCoat(Coat):
         '''
         ck = self.packet.data['ck']
 
-        if ck not in raeting.COAT_KIND_NAMES:
-            self.packet.data['ck'] = raeting.coatKinds.unknown
+        if ck not in list(CoatKind):
+            self.packet.data['ck'] = CoatKind.unknown.value
             emsg = "Unrecognizable packet coat."
             raise raeting.PacketError(emsg)
 
-        if ck == raeting.coatKinds.nacl:
+        if ck == CoatKind.nacl:
             if self.packed:
-                tl = raeting.tailSizes.nacl # nonce length
+                tl = TailSize.nacl.value # nonce length
                 cipher = self.packed[:-tl]
                 nonce = self.packed[-tl:]
                 msg = self.packet.decrypt(cipher, nonce)
@@ -376,7 +388,7 @@ class RxCoat(Coat):
             else:
                 self.packet.body.packed = self.packed
 
-        if ck == raeting.coatKinds.nada:
+        if ck == CoatKind.nada:
             self.packet.body.packed = self.packed
             pass
 
@@ -400,18 +412,18 @@ class TxFoot(Foot):
         '''
         Composes .packed, which is the packed form of this part
         '''
-        self.packed = ''
+        self.packed = b''
         fk = self.packet.data['fk']
 
-        if fk not in raeting.FOOT_KIND_NAMES:
-            self.packet.data['fk'] = raeting.footKinds.unknown
+        if fk not in list(FootKind):
+            self.packet.data['fk'] = FootKind.unknown.value
             emsg = "Unrecognizable packet foot."
             raise raeting.PacketError(emsg)
 
-        if fk == raeting.footKinds.nacl:
-            self.packed = "".rjust(raeting.footSizes.nacl, '\x00')
+        if fk == FootKind.nacl:
+            self.packed = b''.rjust(FootSize.nacl.value, b'\x00')
 
-        elif fk == raeting.footKinds.nada:
+        elif fk == FootKind.nada:
             pass
 
     def sign(self):
@@ -419,15 +431,15 @@ class TxFoot(Foot):
         Compute signature on packet.packed and update packet.packet with signature
         '''
         fk = self.packet.data['fk']
-        if fk not in raeting.FOOT_KIND_NAMES:
-            self.packet.data['fk'] = raeting.footKinds.unknown
+        if fk not in list(FootKind):
+            self.packet.data['fk'] = FootKind.unknown.value
             emsg = "Unrecognizable packet foot."
             raise raeting.PacketError(emsg)
 
-        if fk == raeting.footKinds.nacl:
+        if fk == FootKind.nacl:
             self.packed = self.packet.signature(self.packet.packed)
 
-        elif fk == raeting.footKinds.nada:
+        elif fk == FootKind.nada:
             pass
 
 class RxFoot(Foot):
@@ -440,10 +452,10 @@ class RxFoot(Foot):
         '''
         fk = self.packet.data['fk']
         fl = self.packet.data['fl']
-        self.packed = ''
+        self.packed = b''
 
-        if fk not in raeting.FOOT_KIND_NAMES:
-            self.packet.data['fk'] = raeting.footKinds.unknown
+        if fk not in list(FootKind):
+            self.packet.data['fk'] = FootKind.unknown.value
             emsg = "Unrecognizable packet foot."
             raise raeting.PacketError(emsg)
 
@@ -453,23 +465,23 @@ class RxFoot(Foot):
 
         self.packed = self.packet.packed[self.packet.size - fl:]
 
-        if fk == raeting.footKinds.nacl:
-            if self.size != raeting.footSizes.nacl:
+        if fk == FootKind.nacl:
+            if self.size != FootSize.nacl:
                 emsg = ("Actual foot size '{0}' does not match "
-                    "kind size '{1}'".format(self.size, raeting.footSizes.nacl))
+                    "kind size '{1}'".format(self.size, FootSize.nacl.value))
                 raise raeting.PacketError(emsg)
 
             signature = self.packed
-            blank = "".rjust(raeting.footSizes.nacl, '\x00')
+            blank = b''.rjust(FootSize.nacl.value, b'\x00')
 
             front = self.packet.packed[:self.packet.size - fl]
 
-            msg = "".join([front, blank])
+            msg = front + blank
             if not self.packet.verify(signature, msg):
                 emsg = "Failed verification"
                 raise raeting.PacketError(emsg)
 
-        if fk == raeting.footKinds.nada:
+        if fk == FootKind.nada:
             pass
 
 class Packet(object):
@@ -479,13 +491,13 @@ class Packet(object):
     def __init__(self, stack=None, data=None, kind=None):
         ''' Setup Packet instance. Meta data for a packet. '''
         self.stack = stack
-        self.packed = ''  # packed string
+        self.packed = b''  # packed string
         self.data = odict(raeting.PACKET_DEFAULTS)
         if data:
             self.data.update(data)
         if kind:
-            if kind not in raeting.PCKT_KIND_NAMES:
-                self.data['pk'] = raeting.pcktKinds.unknown
+            if kind not in list(PcktKind):
+                self.data['pk'] = PcktKind.unknown.value
                 emsg = "Unrecognizable packet kind."
                 raise raeting.PacketError(emsg)
             self.data.update(pk=kind)
@@ -532,6 +544,7 @@ class TxPacket(Packet):
     def index(self):
         '''
         Property is transaction tuple (rf, le, re, si, ti, bf,)
+        rf = cf
         '''
         data = self.data
         cf = data['cf']
@@ -552,7 +565,7 @@ class TxPacket(Packet):
         Sign packet with foot
         '''
         self.foot.sign()
-        self.packed = ''.join([self.head.packed,
+        self.packed = b''.join([self.head.packed,
                                self.coat.packed,
                                self.foot.packed])
 
@@ -573,10 +586,25 @@ class TxPacket(Packet):
         self.coat.pack()
         self.foot.pack()
         self.head.pack()
-        self.packed = ''.join([self.head.packed,
+        self.packed = b''.join([self.head.packed,
                                self.coat.packed,
                                self.foot.packed])
 
+    def repack(self):
+        '''
+        Repack the head and resign. Useful if need to change head flag
+        '''
+        self.foot.pack()  # need to pack to reblank it
+        self.head.pack()
+        self.packed = b''.join([self.head.packed,
+                               self.coat.packed,
+                               self.foot.packed])
+
+        self.sign()  # sign updates self.packed
+        if self.size > raeting.UDP_MAX_PACKET_SIZE:
+            emsg = "Packet length of {0}, exceeds max of {1}".format(
+                self.size, raeting.UDP_MAX_PACKET_SIZE)
+            raise raeting.PacketError(emsg)
 
     def pack(self):
         '''
@@ -587,7 +615,6 @@ class TxPacket(Packet):
             emsg = "Packet length of {0}, exceeds max of {1}".format(
                     self.size, raeting.UDP_MAX_PACKET_SIZE)
             raise raeting.PacketError(emsg)
-
         self.sign()
 
 class RxPacket(Packet):
@@ -609,6 +636,7 @@ class RxPacket(Packet):
     def index(self):
         '''
         Property is transaction tuple (rf, le, re, si, ti, bf,)
+        rf = not cf
         '''
         data = self.data
         cf = data['cf']
@@ -741,7 +769,7 @@ class TxTray(Tray):
         self.current = 0
         self.packets = []
         packet = TxPacket(stack=self.stack,
-                          kind=raeting.pcktKinds.message,
+                          kind=PcktKind.message.value,
                           embody=self.body,
                           data=self.data)
 
@@ -758,9 +786,9 @@ class TxTray(Tray):
         Create packeted segments from .packed using headsize footsize
         '''
         extrasize = 0
-        if self.data['hk'] == raeting.headKinds.raet:
+        if self.data['hk'] == HeadKind.raet:
             extrasize = 27 # extra header size as a result of segmentation
-        elif self.data['hk'] == raeting.headKinds.json:
+        elif self.data['hk'] == HeadKind.json:
             extrasize = 36 # extra header size as a result of segmentation
 
         hotelsize = headsize + extrasize + footsize
@@ -779,9 +807,9 @@ class TxTray(Tray):
             packet.coat.packed = packet.body.packed = segment
             packet.foot.pack()
             packet.head.pack()
-            packet.packed = ''.join([  packet.head.packed,
-                                        packet.coat.packed,
-                                        packet.foot.packed])
+            packet.packed = b''.join([packet.head.packed,
+                                     packet.coat.packed,
+                                     packet.foot.packed])
             packet.sign()
             self.packets.append(packet)
 
@@ -810,7 +838,7 @@ class RxTray(Tray):
         console.verbose("segment count={0} number={1} tid={2}\n".format(
             sc, sn, packet.data['ti']))
 
-        if sc == 1:
+        if sc == 1:  # this is only segment to complete now
             self.data.update(packet.data)
             packet.parseInner()
             self.body = packet.body.data
@@ -833,7 +861,8 @@ class RxTray(Tray):
 
     def missing(self, begin=None, end=None):
         '''
-        return list of missing packet numbers between begin and end
+        return deque of missing packet numbers between begin and end where
+        after at least one is found one starting from the end
         '''
         if begin is None:
             begin = 0
@@ -841,11 +870,14 @@ class RxTray(Tray):
             end = len(self.segments)
         if begin >= end:
             return []
-        misseds = []
-        for i, segment in enumerate(self.segments[begin:end]):
-            if segment is None:
-                misseds.append(i)
-        return misseds
+        misseds = deque()
+        found = False  # first missed after found
+        for i in range(end - 1, begin - 1, -1):  # iterate backwards
+            if self.segments[i]:
+                found =  True
+            if found and self.segments[i] is None:
+                misseds.appendleft(i)  # this reverses order
+        return list(misseds)   # convert into a list
 
     def desegmentize(self):
         '''
@@ -853,7 +885,7 @@ class RxTray(Tray):
         and processed header data
         '''
         sc = self.data['sc']
-        self.packed = "".join(self.segments)
+        self.packed = b''.join(self.segments)
         ml = self.data['ml']
         if sc > 1 and self.size != ml:
             emsg = ("Full message payload length '{0}' does not equal head field"
