@@ -29,6 +29,8 @@ from ioflo.base.odicting import odict
 from ioflo.base import aiding
 from ioflo.base import storing
 
+# Import raet libs
+from .abiding import *  # import globals
 from . import raeting
 from . import keeping
 from . import lotting
@@ -274,13 +276,17 @@ class Stack(object):
         Handle one received message from server
         assumes that there is a server
         '''
-        rx, ra = self.server.receive()  # if no data the duple is ('',None)
+        try:
+            rx, ra = self.server.receive()  # if no data the duple is ('',None)
+        except socket.error as ex:
+            if ex.errno == errno.ECONNRESET:
+                return False
         if not rx:  # no received data
             return False
-        # duple = ( packet, source address)
-        self.rxes.append((rx, ra))
+        self.rxes.append((rx, ra))     # duple = ( packet, source address)
         return True
 
+
     def serviceReceives(self):
         '''
         Retrieve from server all recieved and put on the rxes deque
@@ -404,7 +410,8 @@ class Stack(object):
         except socket.error as ex:
             if (ex.errno in [errno.EAGAIN, errno.EWOULDBLOCK,
                              errno.ENETUNREACH, errno.ETIME,
-                             errno.EHOSTUNREACH, errno.EHOSTDOWN]):
+                             errno.EHOSTUNREACH, errno.EHOSTDOWN,
+                             errno.ECONNRESET]):
                 # problem sending such as busy with last message. save it for later
                 laters.append((tx, ta))
                 blocks.append(ta)
